(******************************************************************************)
(* generator (stack machine)                                                  *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Gen;

IMPORT FIO;
IMPORT Comp;

CONST
  debug = FALSE;
  out ::= FIO.StandardOutput;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

  ocMax = 3;      (* 3 operands are enough for l04 (no relations/conditions) *)
  bcMax = ocMax;  (* whole stack may hold accumulator backups *)

TYPE
  TOpdType = ( tVal, tIdx );
  TOpdDesc = RECORD
    CASE t: TOpdType OF
    | tVal: v: Comp.TValue;
    | tIdx: i: Comp.TIndex;
    END;
  END;
  TIdx = [ 1 .. ocMax ];

  TStr = ARRAY [ 0 .. 2 ] OF CHAR;

VAR
  stk: ARRAY TIdx OF TOpdDesc;
  oc: CARDINAL;  (* operand count *)
  vc: CARDINAL;  (* variable count *)
  av: CARDINAL;  (* accumulator value, 0: undef, TIdx: stack *)
  bc: CARDINAL;  (* backup count *)

  opns: ARRAY TOperation OF TStr;


PROCEDURE DbgStk();
VAR i: CARDINAL;
BEGIN
  i := oc;
  WHILE i > 0 DO
    FIO.WrStr( dbg, 'stk[' );
    FIO.WrCard( dbg, i, 0 );
    FIO.WrStr( dbg, '] = ' );
    FIO.WrCard( dbg, ORD(stk[i].t), 0 );
    FIO.WrStr( dbg, ' ' );
    CASE stk[ i ].t OF
    | tVal: FIO.WrInt( dbg, stk[i].v, 11 );
    | tIdx: FIO.WrCard( dbg, stk[i].i, 2 );
    END;
    DEC( i );
    FIO.WrLn( dbg );
  END;
END DbgStk;

PROCEDURE Dbg0();
BEGIN
  FIO.WrStr( dbg, 'Gen.Debug: vc oc av bc' );
  FIO.WrLn( dbg );
END Dbg0;

PROCEDURE DbgS( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( dbg, 'Gen.Debug: ' );
  FIO.WrCard( dbg, vc, 2 );
  FIO.WrStr( dbg, '  ' );
  FIO.WrCard( dbg, oc, 1 );
  FIO.WrStr( dbg, ' ' );
  FIO.WrCard( dbg, av, 2 );
  FIO.WrStr( dbg, '  ' );
  FIO.WrCard( dbg, bc, 1 );
  FIO.WrStr( dbg, ':  ' );
  FIO.WrStr( dbg, s );
END DbgS;

PROCEDURE DbgX();
BEGIN
  FIO.WrLn( dbg );
END DbgX;

PROCEDURE DbgV( v: TValue );
BEGIN
  FIO.WrStr( dbg, '  ' );
  FIO.WrInt( dbg, v, 0 );
  FIO.WrLn( dbg );
END DbgV;

PROCEDURE DbgO( o: TOffset );
BEGIN
  FIO.WrStr( dbg, '   ' );
  FIO.WrCard( dbg, o, 1 );
  FIO.WrLn( dbg );
END DbgO;

PROCEDURE DbgT( x: CARDINAL; s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( dbg, 'Gen.Debug: tmp var ' );
  FIO.WrCard( dbg, x, 0 );
  FIO.WrStr( dbg, ' ' );
  FIO.WrStr( dbg, s );
  FIO.WrLn( dbg );
END DbgT;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Gen: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Value( v: TValue ): Comp.TValue;
BEGIN
  RETURN v;
END Value;

PROCEDURE Index( o: TOffset ): Comp.TIndex;
BEGIN
  INC( o, bcMax );  (* bcMax vars reserved for accumulator backup *)
  IF o >= Comp.cnt THEN Err( 'No such variable' ) END;
  RETURN o;
END Index;

PROCEDURE Lda( i: CARDINAL );
BEGIN
  IF ( i <= 0 ) OR ( i > oc ) THEN Err( 'Stack fault' ) END;
  IF i = av THEN RETURN END;
  IF av > 0 THEN
    IF ( stk[ av ].t = tIdx ) & ( stk[ av ].i < bcMax ) THEN
      IF debug THEN DbgT( stk[av].i, 'updated' ) END;
      Comp.Sto( stk[av].i );
    ELSE
      IF debug THEN DbgT( bc, 'allocated' ) END;
      Comp.Sto( bc );
      stk[ av ].t := tIdx;
      stk[ av ].i := bc;
      INC( bc );
      IF bc > bcMax THEN Err( 'Bug: No backup space left' ) END;
    END;
  END;
  CASE stk[ i ].t OF
  | tVal: Comp.Ldv( stk[i].v );
  | tIdx: Comp.Ldi( stk[i].i );
  END;
END Lda;


PROCEDURE Val( v: TValue );
BEGIN
  IF debug THEN DbgS( 'Val' ); DbgV( v ) END;
  INC( oc );
  IF oc > ocMax THEN Err( 'Stack overflow' ) END;
  stk[ oc ].t := tVal;
  stk[ oc ].v := Value( v );
END Val;

PROCEDURE Lod( o: TOffset );
BEGIN
  IF debug THEN DbgS( 'Lod' ); DbgO( o ) END;
  INC( oc );
  IF oc > ocMax THEN Err( 'Stack overflow' ) END;
  stk[ oc ].t := tIdx;
  stk[ oc ].i := Index( o );
END Lod;

PROCEDURE Chk( c: CARDINAL );
BEGIN
  IF c > oc THEN Err( 'Stack underflow' ) END;
END Chk;

PROCEDURE Dec();
BEGIN
  IF ( stk[ oc ].t = tIdx ) & ( stk[ oc ].i < bcMax ) THEN
    DEC( bc );
    IF debug THEN DbgT( bc, 'released' ) END;
    IF stk[ oc ].i # bc THEN Err( 'Bug: Backup space not a stack :-(' ) END;
  END;
  DEC( oc );
END Dec;

PROCEDURE Sto( o: TOffset );
BEGIN
  IF debug THEN DbgS( 'Sto' ); DbgO( o ) END;
  Chk( 1 ); Lda( oc ); Dec(); Comp.Sto( Index(o) );
  av := 0;
END Sto;

PROCEDURE Calc( o: Comp.TOperation );
BEGIN
  Chk( 2 ); Lda( oc - 1 );
  CASE stk[ oc ].t OF
  | tVal: Comp.Opv( o, stk[oc].v );
  | tIdx: Comp.Opi( o, stk[oc].i );
  END;
  Dec();
  av := oc;
END Calc;

PROCEDURE Opr( o: TOperation );
BEGIN
  IF debug THEN DbgS( opns[o] ); DbgX() END;
  CASE o OF
  | oNeg: Lda( oc ); Comp.Neg(); av := oc;
  | oWrt: Lda( oc ); Dec(); Comp.Out(); av := 0;
  | oAdd: Calc( Comp.oAdd );
  | oSub: Calc( Comp.oSub ); 
  | oMul: Calc( Comp.oMul );
  | oDiv: Calc( Comp.oDiv );
  | oNew: INC( vc ); IF vc > Comp.cnt THEN Err( 'Too many variables' ) END;
  ELSE
    Err( 'Bug: Invalid operation' );
  END;
END Opr;

BEGIN
  oc := 0;
  vc := bcMax;     (* bcMax vars reserved for accumulator backup *)
  av := 0;
  bc := 0;
  opns[ oNeg ] := 'Neg';
  opns[ oWrt ] := 'Wrt';
  opns[ oAdd ] := 'Add';
  opns[ oSub ] := 'Sub';
  opns[ oMul ] := 'Mul';
  opns[ oDiv ] := 'Div';
  opns[ oNew ] := 'New';
  IF debug THEN Dbg0() END;
END Gen.
